#
#   Imandra Inc.
#
#   codelogician/commands/cmd_multiagent.py
#

from typing import Annotated

import typer

from codelogician.util import setup_logging


# fmt: off
def run_multiagent_cmd(
    dir   : Annotated[str, typer.Argument(help='Target directory')],
    clean : Annotated[bool, typer.Option(help='Start clean by disregarding any existing cache files')] = False,
    config: Annotated[str, typer.Option(help='Server configuration YAML file')] = 'config/server_config.yaml',
    debug : Annotated[bool, typer.Option(help='Debug mode')] = False,
):
    from codelogician.server.oneshot import run_oneshot

    setup_logging(debug=debug)
    run_oneshot(dir, clean, config)
# fmt: on
